# cspell:ignore rpartition
# pylint: disable=redefined-builtin
"""Create interactive plots for `sympy` expressions.

The procedure to create interactive plots with for :mod:`sympy` expressions
with :doc:`mpl-interactions <mpl_interactions:index>` has been extracted to
this module.

The module is only available here, under the documentation. If this feature
turns out to be popular, it can be published as an independent package.

The package also provides other helpful functions, like
:func:`substitute_indexed_symbols`, that are useful when visualizing
`sympy` expressions.
"""
from __future__ import annotations

import inspect
import logging
import sys
from collections import abc
from typing import (
    TYPE_CHECKING,
    Callable,
    Iterator,
    Mapping,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import sympy as sp
from ipywidgets.widgets import FloatSlider, IntSlider
from sympy.printing.latex import translate

if TYPE_CHECKING:  # pragma: no cover
    from IPython.lib.pretty import PrettyPrinter

    if sys.version_info >= (3, 10):
        from typing import TypeGuard  # pylint: disable=no-name-in-module
    else:
        from typing_extensions import TypeGuard


Slider = Union[FloatSlider, IntSlider]
"""Allowed :doc:`ipywidgets <ipywidgets:index>` slider types."""
RangeDefinition = Union[
    Tuple[float, float],
    Tuple[float, float, Union[float, int]],
]
"""Types of range definitions used in :meth:`.set_ranges`."""


class SliderKwargs(abc.Mapping):
    """Wrapper around a `dict` of sliders that can serve as keyword arguments.

    Sliders can be defined in :func:`~mpl_interactions.pyplot.interactive_plot`
    through :term:`kwargs <python:keyword argument>`. This wrapper class can be
    used for that.

    .. automethod:: __getitem__
    """

    def __init__(
        self,
        sliders: Mapping[str, Slider],
        arg_to_symbol: Mapping[str, str],
    ) -> None:
        self._verify_arguments(sliders, arg_to_symbol)
        self._sliders = dict(sliders)
        self._arg_to_symbol = {
            arg: symbol
            for arg, symbol in arg_to_symbol.items()
            if symbol in self._sliders
        }

    @property
    def arg_to_symbol(self) -> dict[str, str]:
        """**Copy** of the internal translation `dict` for argument names."""
        return dict(self._arg_to_symbol)

    @property
    def symbol_to_arg(self) -> dict[str, str]:
        """Inverted `dict` of `arg_to_symbol`."""
        return {symbol: arg for arg, symbol in self._arg_to_symbol.items()}

    @staticmethod
    def _verify_arguments(
        sliders: Mapping[str, Slider], arg_to_symbol: Mapping[str, str]
    ) -> None:
        symbol_names = set(arg_to_symbol.values())
        for arg_name in arg_to_symbol:
            if not arg_name.isidentifier():
                raise ValueError(
                    f'Argument "{arg_name}" in arg_to_symbol is not a'
                    " valid identifier for a Python variable"
                )
        for slider_name in sliders:
            if not isinstance(slider_name, str):
                raise TypeError(
                    f'Slider name "{slider_name}" is not of type str'
                )
            if slider_name not in symbol_names:
                raise ValueError(
                    f'Slider with name "{slider_name}" is not covered by '
                    "arg_to_symbol"
                )
        for name, slider in sliders.items():
            if not isinstance(slider, Slider.__args__):  # type: ignore[attr-defined]
                raise TypeError(
                    f'Slider "{name}" is not a valid ipywidgets slider'
                )

    def __getitem__(self, key: str | sp.Symbol) -> Slider:
        """Get slider by symbol, symbol name, or argument name."""
        if isinstance(key, sp.Symbol):
            key = key.name
        if key in self._arg_to_symbol:
            slider_name = self._arg_to_symbol[key]
        else:
            slider_name = key
        if slider_name not in self._sliders:
            raise KeyError(f'"{key}" is neither an argument nor a symbol name')
        return self._sliders[slider_name]

    def __iter__(self) -> Iterator[str]:
        """Iterate over the arguments of the `.LambdifiedExpression`.

        This is useful for unpacking an instance of `SliderKwargs` as
        :term:`kwargs <python:keyword argument>`.
        """
        return self._arg_to_symbol.__iter__()

    def __len__(self) -> int:
        return len(self._sliders)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"sliders={self._sliders}, "
            f"arg_to_symbol={self._arg_to_symbol})"
        )

    def _repr_pretty_(self, p: PrettyPrinter, cycle: bool) -> None:
        class_name = type(self).__name__
        if cycle:
            p.text(f"{class_name}(...)")
        else:
            with p.group(indent=2, open=f"{class_name}("):
                p.breakable()
                p.text("sliders=")
                p.pretty(self._sliders)
                p.text(",")
                p.breakable()
                p.text("arg_to_symbol=")
                p.pretty(self._arg_to_symbol)
                p.text(",")
            p.breakable()
            p.text(")")

    def set_values(self, *args: dict[str, float], **kwargs: float) -> None:
        """Set initial values for the sliders.

        Either use a `dict` as input, or use :term:`kwargs <python:keyword
        argument>` with slider names as the keywords (see
        `.SliderKwargs.__getitem__`). This façade method exists in particular
        for `.parameter_defaults`.
        """
        value_mapping = _merge_args_kwargs(*args, **kwargs)
        for keyword, value in value_mapping.items():
            try:
                self[keyword].value = value
            except KeyError:
                logging.warning(
                    f'There is no slider with name or symbol "{keyword}"'
                )
                continue

    def set_ranges(  # noqa: R701
        self, *args: dict[str, RangeDefinition], **kwargs: RangeDefinition
    ) -> None:
        """Set min, max and (optionally) the nr of steps for each slider.

        .. tip::
            :code:`n_steps` becomes the step **size** if its value is
            `float`.
        """
        range_definitions = _merge_args_kwargs(*args, **kwargs)
        for slider_name, range_def in range_definitions.items():
            if not isinstance(range_def, tuple):
                raise TypeError(
                    f'Range definition for slider "{slider_name}" is not a'
                    " tuple"
                )
            slider = self[slider_name]
            if _is_min_max(range_def):
                min_, max_ = range_def
                step_size = slider.step
            elif _is_min_max_step(range_def):
                min_, max_, n_steps = range_def
                if n_steps <= 0:
                    raise ValueError("Number of steps has to be positive")
                if isinstance(n_steps, float):
                    step_size = n_steps
                else:
                    step_size = (max_ - min_) / n_steps
            else:
                raise ValueError(
                    f'Range definition {range_def} for slider "{slider_name}"'
                    " is neither of shape (min, max) nor (min, max, n_steps)"
                )
            if min_ > slider.max:
                slider.max = max_
                slider.min = min_
            else:
                slider.min = min_
                slider.max = max_
            if isinstance(slider, FloatSlider):
                slider.step = step_size


def _is_min_max(
    range_def: RangeDefinition,
) -> TypeGuard[tuple[float, float]]:
    if len(range_def) == 2:
        return True
    return False


def _is_min_max_step(
    range_def: RangeDefinition,
) -> TypeGuard[tuple[float, float, float | int]]:
    if len(range_def) == 3:
        return True
    return False


ValueType = TypeVar("ValueType")


def _merge_args_kwargs(
    *args: dict[str, ValueType], **kwargs: ValueType
) -> dict[str, ValueType]:
    r"""Merge positional `dict` arguments and keyword arguments into one `dict`.

    >>> _merge_args_kwargs(x="X", y="Y")
    {'x': 'X', 'y': 'Y'}
    >>> _merge_args_kwargs({R"\theta": 0}, a=1, b=2)
    {'\\theta': 0, 'a': 1, 'b': 2}
    """
    output_dict = {}
    for arg in args:
        if not isinstance(arg, abc.Mapping):
            raise TypeError("Positional arguments have to be of type dict")
        output_dict.update(arg)
    output_dict.update(kwargs)
    return output_dict


def prepare_sliders(
    expression: sp.Expr, plot_symbol: sp.Symbol | tuple[sp.Symbol, ...]
) -> tuple[Callable, SliderKwargs]:
    # cspell:ignore lambdifygenerated
    """Lambdify a `sympy` expression and create sliders for its arguments.

    >>> n = sp.Symbol("n", integer=True)
    >>> x = sp.Symbol("x")
    >>> expression, sliders = prepare_sliders(x ** n, plot_symbol=x)
    >>> expression
    <function _lambdifygenerated at ...>
    >>> sliders
    SliderKwargs(...)
    """
    plot_symbols = __safe_wrap_symbols(plot_symbol)
    slider_symbols = _extract_slider_symbols(expression, plot_symbols)
    lambdified_expression = sp.lambdify(
        (*plot_symbols, *slider_symbols),
        expression,
        modules="numpy",
    )
    sliders_mapping = {
        symbol.name: create_slider(symbol) for symbol in slider_symbols
    }
    symbols_names = map(lambda s: s.name, (*plot_symbols, *slider_symbols))
    arg_names = inspect.signature(lambdified_expression).parameters
    arg_to_symbol = dict(zip(arg_names, symbols_names))
    sliders = SliderKwargs(sliders_mapping, arg_to_symbol)
    return lambdified_expression, sliders


def create_slider(symbol: sp.Symbol) -> Slider:
    r"""Create an `int` or `float` slider, depending on Symbol assumptions.

    The description for the slider is rendered as LaTeX from the
    `~sympy.core.symbol.Symbol` name.

    >>> create_slider(sp.Symbol("a"))
    FloatSlider(value=0.0, description='\\(a\\)')
    >>> create_slider(sp.Symbol("n0", integer=True))
    IntSlider(value=0, description='\\(n_{0}\\)')
    """
    description = Rf"\({sp.latex(symbol)}\)"
    if symbol.is_integer:  # type: ignore[attr-defined]
        return IntSlider(description=description)
    return FloatSlider(description=description)


def _extract_slider_symbols(
    expression: sp.Expr,
    plot_symbol: sp.Symbol | Sequence[sp.Symbol],
) -> tuple[sp.Symbol, ...]:
    """Extract sorted, remaining free symbols of a `sympy` expression."""
    plot_symbols = __safe_wrap_symbols(plot_symbol)
    free_symbols: set[sp.Symbol] = expression.free_symbols  # type: ignore[assignment]
    for symbol in plot_symbols:
        if symbol not in free_symbols:
            raise ValueError(
                f"Expression does not contain a free symbol named {symbol}"
            )
        free_symbols.remove(symbol)
    ordered_symbols = sorted(free_symbols, key=lambda s: s.name)
    return tuple(ordered_symbols)


def __safe_wrap_symbols(
    plot_symbol: sp.Symbol | Sequence[sp.Symbol],
) -> tuple[sp.Symbol, ...]:
    if isinstance(plot_symbol, abc.Sequence):
        return tuple(plot_symbol)
    if isinstance(plot_symbol, sp.Symbol):
        return (plot_symbol,)
    raise TypeError(
        f"Wrong plot_symbol input type {type(plot_symbol).__name__}"
    )


def partial_doit(
    expression: sp.Expr,
    doit_classes: type[sp.Basic] | tuple[type[sp.Basic], ...],
) -> sp.Expr:
    """Perform :meth:`~sympy.core.basic.Basic.doit` up to a certain level.

    Arguments
    ---------
    expression: the `~sympy.core.expr.Expr` on which you want to perform a
        :meth:`~sympy.core.basic.Basic.doit`.
    doit_classes: types on which the :meth:`~sympy.core.basic.Basic.doit`
        should be performed.
    """
    new_expression = expression
    for node in sp.preorder_traversal(expression):
        if isinstance(node, doit_classes):
            new_expression = new_expression.xreplace(
                {node: node.doit(deep=False)}
            )
    return new_expression


def _indexed_to_symbol(idx: sp.Indexed) -> sp.Symbol:
    base_name, _, _ = str(idx).rpartition("[")
    subscript = ",".join(map(str, idx.indices))
    if len(idx.indices) > 1:
        base_name = translate(base_name)
        subscript = "_{" + subscript + "}"
    return sp.Symbol(f"{base_name}{subscript}")


def rename_symbols(
    expression: sp.Expr, renames: Callable[[str], str] | dict[str, str]
) -> sp.Expr:
    r"""Rename symbols in an expression.

    >>> a, b, x = sp.symbols(R"a \beta x")
    >>> expr = a + b * x
    >>> rename_symbols(expr, renames={"a": "A", R"\beta": "B"})
    A + B*x
    >>> rename_symbols(expr, renames=lambda s: s.replace("\\", ""))
    a + beta*x
    >>> rename_symbols(expr, renames={"non-existent": "c"})
    Traceback (most recent call last):
        ...
    KeyError: "No symbol with name 'non-existent' in expression"
    """
    substitutions: dict[sp.Symbol, sp.Symbol] = {}
    free_symbols: set[sp.Symbol] = expression.free_symbols  # type: ignore[assignment]
    if callable(renames):
        for old_symbol in free_symbols:
            new_name = renames(old_symbol.name)
            new_symbol = sp.Symbol(new_name, **old_symbol.assumptions0)
            substitutions[old_symbol] = new_symbol
    elif isinstance(renames, dict):
        for old_name, new_name in renames.items():
            # pylint: disable=cell-var-from-loop
            matches = filter(lambda s: s.name == old_name, free_symbols)
            try:
                old_symbol = next(matches)
            except StopIteration:
                # pylint: disable=raise-missing-from
                raise KeyError(
                    f"No symbol with name '{old_name}' in expression"
                )
            new_symbol = sp.Symbol(new_name, **old_symbol.assumptions0)
            substitutions[old_symbol] = new_symbol
    else:
        raise TypeError(f"Cannot rename from type {type(renames).__name__}")
    return expression.xreplace(substitutions)


def substitute_indexed_symbols(expression: sp.Expr) -> sp.Expr:
    """Substitute `~sympy.tensor.indexed.IndexedBase` with symbols.

    See :doc:`compwa-org:report/008` for more info.
    """
    return expression.xreplace(
        {
            s: _indexed_to_symbol(s)
            for s in expression.free_symbols
            if isinstance(s, sp.Indexed)
        }
    )



# cspell:ignore Asner mhash
# pylint: disable=arguments-differ,protected-access,unbalanced-tuple-unpacking,
# pylint: disable=unused-argument, W0223
# https://stackoverflow.com/a/22224042
"""Lineshape functions that describe the dynamics of an interaction.

.. seealso:: :doc:`/usage/dynamics` and
    :doc:`/usage/dynamics/analytic-continuation`
"""
from __future__ import annotations

import re
import sys
from typing import Sequence

import sympy as sp
from sympy.printing.conventions import split_super_sub
from sympy.printing.latex import LatexPrinter

from ampform.sympy import (
    UnevaluatedExpression,
    create_expression,
    implement_doit_method,
)
from ampform.sympy.math import ComplexSqrt

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol  # pragma: no cover


@implement_doit_method
class BlattWeisskopfSquared(UnevaluatedExpression):
    # cspell:ignore pychyGekoppeltePartialwellenanalyseAnnihilationen
    r"""Blatt-Weisskopf function :math:`B_L^2(z)`, up to :math:`L \leq 8`.

    Args:
        angular_momentum: Angular momentum :math:`L` of the decaying particle.

        z: Argument of the Blatt-Weisskopf function :math:`B_L^2(z)`. A usual
            choice is :math:`z = (d q)^2` with :math:`d` the impact parameter
            and :math:`q` the breakup-momentum (see
            :func:`BreakupMomentumSquared`).

    Note that equal powers of :math:`z` appear in the nominator and the
    denominator, while some sources have nominator :math:`1`, instead of
    :math:`z^L`. Compare for instance :pdg-review:`2020; Resonances; p.6`, just
    before Equation (49.20).

    Each of these cases for :math:`L` has been taken from
    :cite:`pychyGekoppeltePartialwellenanalyseAnnihilationen2016`, p.59,
    :cite:`chungPartialWaveAnalysis1995`, p.415, and
    :cite:`chungFormulasAngularMomentumBarrier2015`. For a good overview of
    where to use these Blatt-Weisskopf functions, see
    :cite:`asnerDalitzPlotAnalysis2006`.

    See also :ref:`usage/dynamics:Form factor`.
    """
    is_commutative = True
    max_angular_momentum: int | None = None
    """Limit the maximum allowed angular momentum :math:`L`.

    This improves performance when :math:`L` is a `~sympy.core.symbol.Symbol`
    and you are note interested in higher angular momenta.
    """

    def __new__(cls, angular_momentum, z, **hints) -> BlattWeisskopfSquared:
        return create_expression(cls, angular_momentum, z, **hints)

    def evaluate(self) -> sp.Expr:
        angular_momentum: sp.Expr = self.args[0]  # type: ignore[assignment]
        z: sp.Expr = self.args[1]  # type: ignore[assignment]
        cases: dict[int, sp.Expr] = {
            0: sp.S.One,
            1: 2 * z / (z + 1),
            2: 13 * z**2 / ((z - 3) * (z - 3) + 9 * z),
            3: (
                277
                * z**3
                / (z * (z - 15) * (z - 15) + 9 * (2 * z - 5) * (2 * z - 5))
            ),
            4: (
                12746
                * z**4
                / (
                    (z**2 - 45 * z + 105) * (z**2 - 45 * z + 105)
                    + 25 * z * (2 * z - 21) * (2 * z - 21)
                )
            ),
            5: 998881
            * z**5
            / (
                z**5
                + 15 * z**4
                + 315 * z**3
                + 6300 * z**2
                + 99225 * z
                + 893025
            ),
            6: 118394977
            * z**6
            / (
                z**6
                + 21 * z**5
                + 630 * z**4
                + 18900 * z**3
                + 496125 * z**2
                + 9823275 * z
                + 108056025
            ),
            7: 19727003738
            * z**7
            / (
                z**7
                + 28 * z**6
                + 1134 * z**5
                + 47250 * z**4
                + 1819125 * z**3
                + 58939650 * z**2
                + 1404728325 * z
                + 18261468225
            ),
            8: 4392846440677
            * z**8
            / (
                z**8
                + 36 * z**7
                + 1890 * z**6
                + 103950 * z**5
                + 5457375 * z**4
                + 255405150 * z**3
                + 9833098275 * z**2
                + 273922023375 * z
                + 4108830350625
            ),
        }
        return sp.Piecewise(
            *[
                (expression, sp.Eq(angular_momentum, value))
                for value, expression in cases.items()
                if self.max_angular_momentum is None
                or value <= self.max_angular_momentum
            ]
        )

    def _latex(self, printer: LatexPrinter, *args) -> str:
        angular_momentum, z = tuple(map(printer._print, self.args))
        return Rf"B_{{{angular_momentum}}}^2\left({z}\right)"


class PhaseSpaceFactorProtocol(Protocol):
    """Protocol that is used by `.EnergyDependentWidth`.

    Use this `~typing.Protocol` when defining other implementations of a phase
    space factor. Compare for instance `.PhaseSpaceFactor` and
    `.PhaseSpaceFactorAnalytic`.
    """

    def __call__(self, s, m_a, m_b) -> sp.Expr:
        """Expected `~inspect.signature`."""
        ...  # pragma: no cover


@implement_doit_method
class PhaseSpaceFactor(UnevaluatedExpression):
    """Standard phase-space factor, using :func:`BreakupMomentumSquared`.

    See :pdg-review:`2020; Resonances; p.4`, Equation (49.8).
    """

    is_commutative = True

    def __new__(cls, s, m_a, m_b, **hints) -> PhaseSpaceFactor:
        return create_expression(cls, s, m_a, m_b, **hints)

    def evaluate(self) -> sp.Expr:
        s, m_a, m_b = self.args
        q_squared = BreakupMomentumSquared(s, m_a, m_b)
        denominator = _phase_space_factor_denominator(s)
        return sp.sqrt(q_squared) / denominator

    def _latex(self, printer: LatexPrinter, *args) -> str:
        s, *_ = self.args
        s = printer._print(s)
        subscript = _indices_to_subscript(_determine_indices(s))
        name = R"\rho" + subscript if self._name is None else self._name
        return Rf"{name}\left({s}\right)"


@implement_doit_method
class PhaseSpaceFactorAbs(UnevaluatedExpression):
    r"""Phase space factor square root over the absolute value.

    As opposed to `.PhaseSpaceFactor`, this takes the
    `~sympy.functions.elementary.complexes.Abs` value of
    `.BreakupMomentumSquared`, then the
    :func:`~sympy.functions.elementary.miscellaneous.sqrt`.

    This version of the phase space factor is often denoted as
    :math:`\hat{\rho}` and is used in analytic continuation
    (`.PhaseSpaceFactorAnalytic`).
    """

    is_commutative = True

    def __new__(cls, s, m_a, m_b, **hints) -> PhaseSpaceFactorAbs:
        return create_expression(cls, s, m_a, m_b, **hints)

    def evaluate(self) -> sp.Expr:
        s, m_a, m_b = self.args
        q_squared = BreakupMomentumSquared(s, m_a, m_b)
        denominator = _phase_space_factor_denominator(s)
        return sp.sqrt(sp.Abs(q_squared)) / denominator

    def _latex(self, printer: LatexPrinter, *args) -> str:
        s, *_ = self.args
        s = printer._print(s)
        subscript = _indices_to_subscript(_determine_indices(s))
        name = R"\hat{\rho}" + subscript if self._name is None else self._name
        return Rf"{name}\left({s}\right)"


@implement_doit_method
class PhaseSpaceFactorAnalytic(UnevaluatedExpression):
    """Analytic continuation for the :func:`PhaseSpaceFactor`.

    See :pdg-review:`2018; Resonances; p.9` and
    :doc:`/usage/dynamics/analytic-continuation`.

    **Warning**: The PDG specifically derives this formula for a two-body decay
    *with equal masses*.
    """

    is_commutative = True

    def __new__(cls, s, m_a, m_b, **hints) -> PhaseSpaceFactorAnalytic:
        return create_expression(cls, s, m_a, m_b, **hints)

    def evaluate(self) -> sp.Expr:
        s, m_a, m_b = self.args
        rho_hat = PhaseSpaceFactorAbs(s, m_a, m_b)
        s_threshold = (m_a + m_b) ** 2  # type: ignore[operator]
        return _analytic_continuation(rho_hat, s, s_threshold)

    def _latex(self, printer: LatexPrinter, *args) -> str:
        s, *_ = self.args
        s = printer._print(s)
        subscript = _indices_to_subscript(_determine_indices(s))
        name = (
            R"\rho^\mathrm{ac}" + subscript
            if self._name is None
            else self._name
        )
        return Rf"{name}\left({s}\right)"


@implement_doit_method
class PhaseSpaceFactorComplex(UnevaluatedExpression):
    """Phase-space factor with `.ComplexSqrt`.

    Same as :func:`PhaseSpaceFactor`, but using a `.ComplexSqrt` that does have
    defined behavior for defined for negative input values.
    """

    is_commutative = True

    def __new__(cls, s, m_a, m_b, **hints) -> PhaseSpaceFactorComplex:
        return create_expression(cls, s, m_a, m_b, **hints)

    def evaluate(self) -> sp.Expr:
        s, m_a, m_b = self.args
        q_squared = BreakupMomentumSquared(s, m_a, m_b)
        denominator = _phase_space_factor_denominator(s)
        return ComplexSqrt(q_squared) / denominator

    def _latex(self, printer: LatexPrinter, *args) -> str:
        s, *_ = self.args
        s = printer._print(s)
        subscript = _indices_to_subscript(_determine_indices(s))
        name = (
            R"\rho^\mathrm{c}" + subscript
            if self._name is None
            else self._name
        )
        return Rf"{name}\left({s}\right)"


def _analytic_continuation(rho, s, s_threshold) -> sp.Piecewise:
    return sp.Piecewise(
        (
            sp.I * rho / sp.pi * sp.log(sp.Abs((1 + rho) / (1 - rho))),
            s < 0,
        ),
        (
            rho + sp.I * rho / sp.pi * sp.log(sp.Abs((1 + rho) / (1 - rho))),
            s > s_threshold,
        ),
        (
            2 * sp.I * rho / sp.pi * sp.atan(1 / rho),
            True,
        ),
    )


def _phase_space_factor_denominator(s) -> sp.Mul:
    return 8 * sp.pi * sp.sqrt(s)


@implement_doit_method
class EnergyDependentWidth(UnevaluatedExpression):
    r"""Mass-dependent width, coupled to the pole position of the resonance.

    See :pdg-review:`2020; Resonances; p.6` and
    :cite:`asnerDalitzPlotAnalysis2006`, equation (6). Default value for
    :code:`phsp_factor` is :meth:`PhaseSpaceFactor`.

    Note that the `.BlattWeisskopfSquared` of AmpForm is normalized in the
    sense that equal powers of :math:`z` appear in the nominator and the
    denominator, while the definition in the PDG (as well as some other
    sources), always have :math:`1` in the nominator of the Blatt-Weisskopf. In
    that case, one needs an additional factor :math:`\left(q/q_0\right)^{2L}`
    in the definition for :math:`\Gamma(m)`.
    """

    # https://github.com/sympy/sympy/blob/1.8/sympy/core/basic.py#L74-L77
    __slots__ = ("phsp_factor",)
    phsp_factor: PhaseSpaceFactorProtocol
    is_commutative = True

    def __new__(  # pylint: disable=too-many-arguments
        cls,
        s,
        mass0,
        gamma0,
        m_a,
        m_b,
        angular_momentum,
        meson_radius,
        phsp_factor: PhaseSpaceFactorProtocol | None = None,
        name: str | None = None,
        evaluate: bool = False,
    ) -> EnergyDependentWidth:
        args = sp.sympify(
            (s, mass0, gamma0, m_a, m_b, angular_momentum, meson_radius)
        )
        if phsp_factor is None:
            phsp_factor = PhaseSpaceFactor
        # Overwritting Basic.__new__ to store phase space factor type
        # https://github.com/sympy/sympy/blob/1.8/sympy/core/basic.py#L113-L119
        expr = object.__new__(cls)
        expr._assumptions = cls.default_assumptions  # type: ignore[attr-defined]
        expr._mhash = None
        expr._args = args
        expr._name = name
        expr.phsp_factor = phsp_factor
        if evaluate:
            return expr.evaluate()  # type: ignore[return-value]
        return expr

    def __getnewargs__(self) -> tuple:
        # Pickling support, see
        # https://github.com/sympy/sympy/blob/1.8/sympy/core/basic.py#L124-L126
        return (*self.args, self.phsp_factor, self._name)

    def evaluate(self) -> sp.Expr:
        s, mass0, gamma0, m_a, m_b, angular_momentum, meson_radius = self.args
        q_squared = BreakupMomentumSquared(s, m_a, m_b)
        q0_squared = BreakupMomentumSquared(mass0**2, m_a, m_b)  # type: ignore[operator]
        form_factor_sq = BlattWeisskopfSquared(
            angular_momentum,
            z=q_squared * meson_radius**2,  # type: ignore[operator]
        )
        form_factor0_sq = BlattWeisskopfSquared(
            angular_momentum,
            z=q0_squared * meson_radius**2,  # type: ignore[operator]
        )
        rho = self.phsp_factor(s, m_a, m_b)
        rho0 = self.phsp_factor(mass0**2, m_a, m_b)  # type: ignore[operator]
        return gamma0 * (form_factor_sq / form_factor0_sq) * (rho / rho0)

    def _latex(self, printer: LatexPrinter, *args) -> str:
        s, _, width, *_ = self.args
        s = printer._print(s)
        subscript = _indices_to_subscript(_determine_indices(width))
        name = Rf"\Gamma{subscript}" if self._name is None else self._name
        return Rf"{name}\left({s}\right)"


@implement_doit_method
class BreakupMomentumSquared(UnevaluatedExpression):
    r"""Squared value of the two-body break-up momentum.

    For a two-body decay :math:`R \to ab`, the *break-up momentum* is the
    absolute value of the momentum of both :math:`a` and :math:`b` in the rest
    frame of :math:`R`.

    Args:
        s: :ref:`Mandelstam variable <pwa:mandelstam-variables>` :math:`s`.
            Commonly, this is just :math:`s = m_R^2`, with :math:`m_R` the
            invariant mass of decaying particle :math:`R`.

        m_a: Mass of decay product :math:`a`.
        m_b: Mass of decay product :math:`b`.

    It's up to the caller in which way to take the square root of this break-up
    momentum.See :doc:`/usage/dynamics/analytic-continuation` and
    `.ComplexSqrt`.
    """

    is_commutative = True

    def __new__(cls, s, m_a, m_b, **hints) -> BreakupMomentumSquared:
        return create_expression(cls, s, m_a, m_b, **hints)

    def evaluate(self) -> sp.Expr:
        s, m_a, m_b = self.args
        return (s - (m_a + m_b) ** 2) * (s - (m_a - m_b) ** 2) / (4 * s)  # type: ignore[operator]

    def _latex(self, printer: LatexPrinter, *args) -> str:
        s, *_ = self.args
        s = printer._print(s)
        subscript = _indices_to_subscript(_determine_indices(s))
        name = "q^2" + subscript if self._name is None else self._name
        return Rf"{name}\left({s}\right)"


def relativistic_breit_wigner(s, mass0, gamma0) -> sp.Expr:
    """Relativistic Breit-Wigner lineshape.

    See :ref:`usage/dynamics:_Without_ form factor` and
    :cite:`asnerDalitzPlotAnalysis2006`.
    """
    return gamma0 * mass0 / (mass0**2 - s - gamma0 * mass0 * sp.I)


def relativistic_breit_wigner_with_ff(  # pylint: disable=too-many-arguments
    s,
    mass0,
    gamma0,
    m_a,
    m_b,
    angular_momentum,
    meson_radius,
    phsp_factor: PhaseSpaceFactorProtocol | None = None,
) -> sp.Expr:
    """Relativistic Breit-Wigner with `.BlattWeisskopfSquared` factor.

    See :ref:`usage/dynamics:_With_ form factor` and
    :pdg-review:`2020; Resonances; p.6`.
    """
    q_squared = BreakupMomentumSquared(s, m_a, m_b)
    ff_squared = BlattWeisskopfSquared(
        angular_momentum, z=q_squared * meson_radius**2
    )
    form_factor = sp.sqrt(ff_squared)
    mass_dependent_width = EnergyDependentWidth(
        s, mass0, gamma0, m_a, m_b, angular_momentum, meson_radius, phsp_factor
    )
    return (mass0 * gamma0 * form_factor) / (
        mass0**2 - s - mass_dependent_width * mass0 * sp.I
    )


def _indices_to_subscript(indices: Sequence[int]) -> str:
    """Create a LaTeX subscript from a list of indices.

    >>> _indices_to_subscript([])
    ''
    >>> _indices_to_subscript([1])
    '_{1}'
    >>> _indices_to_subscript([1, 2])
    '_{1,2}'
    """
    if len(indices) == 0:
        return ""
    subscript = ",".join(map(str, indices))
    return "_{" + subscript + "}"


def _determine_indices(symbol) -> list[int]:
    r"""Extract any indices if available from a `~sympy.core.symbol.Symbol`.

    >>> _determine_indices(sp.Symbol("m1"))
    [1]
    >>> _determine_indices(sp.Symbol("m_a2"))
    [2]
    >>> _determine_indices(sp.Symbol(R"\alpha_{i2, 5}"))
    [2, 5]
    >>> _determine_indices(sp.Symbol("m"))
    []

    `~sympy.tensor.indexed.Indexed` instances can also be handled:
    >>> m_a = sp.IndexedBase("m_a")
    >>> _determine_indices(m_a[0])
    [0]
    """
    _, _, subscripts = split_super_sub(sp.latex(symbol))
    if not subscripts:
        return []
    subscript: str = subscripts[-1]
    subscript = re.sub(r"[^0-9^\,]", "", subscript)
    subscript = f"[{subscript}]"
    try:
        indices = eval(subscript)  # pylint: disable=eval-used
    except SyntaxError:
        return []
    return list(indices)



from __future__ import annotations

import pytest
from qrules import ReactionInfo

from ampform import get_builder
from ampform.helicity import HelicityModel
from ampform.helicity.naming import (
    CanonicalAmplitudeNameGenerator,
    HelicityAmplitudeNameGenerator,
    _render_float,
    generate_transition_label,
)


def test_generate_transition_label(reaction: ReactionInfo):
    for transition in reaction.transitions:
        label = generate_transition_label(transition)
        jpsi_spin = _render_float(transition.states[-1].spin_projection)
        gamma_spin = _render_float(transition.states[0].spin_projection)
        assert label == (
            Rf"J/\psi(1S)_{{{jpsi_spin}}} \to \gamma_{{{gamma_spin}}}"
            R" \pi^{0}_{0} \pi^{0}_{0}"
        )


@pytest.mark.parametrize("parent_helicities", [False, True])
@pytest.mark.parametrize("child_helicities", [False, True])
@pytest.mark.parametrize("ls_combinations", [False, True])
def test_coefficient_names(  # noqa: R701
    reaction: ReactionInfo,
    parent_helicities,
    child_helicities,
    ls_combinations,
):
    # pylint: disable=too-many-branches, too-many-statements
    builder = get_builder(reaction)
    assert isinstance(builder.naming, HelicityAmplitudeNameGenerator)
    builder.naming.insert_parent_helicities = parent_helicities
    builder.naming.insert_child_helicities = child_helicities
    if ls_combinations:
        if reaction.formalism == "helicity":
            pytest.skip("No LS-combinations if using helicity formalism")
    if isinstance(builder.naming, CanonicalAmplitudeNameGenerator):
        builder.naming.insert_ls_combinations = ls_combinations
    model = builder.formulate()

    coefficients = get_coefficients(model)
    n_resonances = len(reaction.get_intermediate_particles())
    if reaction.formalism == "helicity":
        if parent_helicities:
            if child_helicities:
                assert len(coefficients) == 4 * n_resonances
            else:
                assert len(coefficients) == 2 * n_resonances
        else:
            if child_helicities:
                assert len(coefficients) == n_resonances
            else:
                assert len(coefficients) == n_resonances
    elif reaction.formalism == "canonical-helicity":
        if ls_combinations:
            if parent_helicities:
                if child_helicities:
                    assert len(coefficients) == 8 * n_resonances
                else:
                    assert len(coefficients) == 4 * n_resonances
            else:
                if child_helicities:
                    assert len(coefficients) == 4 * n_resonances
                else:
                    assert len(coefficients) == 2 * n_resonances
        else:
            if parent_helicities:
                if child_helicities:
                    assert len(coefficients) == 4 * n_resonances
                else:
                    assert len(coefficients) == 2 * n_resonances
            else:
                assert len(coefficients) == n_resonances

    coefficient_name = coefficients[0]
    if parent_helicities:
        assert R"J/\psi(1S)_{-1}" in coefficient_name
    else:
        assert R"J/\psi(1S) " in coefficient_name

    if child_helicities:
        assert R"\gamma_{" in coefficient_name
    else:
        assert R"\gamma;" in coefficient_name

    if ls_combinations:
        assert R"\xrightarrow[S=1]{L=0}" in coefficient_name
    else:
        assert R"\to" in coefficient_name


def get_coefficients(model: HelicityModel) -> list[str]:
    return [
        symbol.name
        for symbol in model.parameter_defaults
        if symbol.name.startswith("C_")
    ]
